/// \file spiuart.c VS1000 UART Audio Module controlling software
/// \author Panu-Kristian Poiksalo, VLSI Solution Oy 2008


#include <stdio.h> //Standard io
#include <ctype.h>
//#include <stdlib.h> // VS_DSP Standard Library
#include <vs1000.h> // VS1000B register definitions
//#include <vectors.h> // VS1000B vectors (interrupts and services)
#include <fat.h> // Read Only Fat Filesystem
#include <string.h> // memcpy etc
#include <player.h> // VS1000B default ROM player
#include <audio.h> // DAC output
//#include <codec.h> // CODEC interface
#include <mapper.h>

#include "uartRxInt.h"
#include "spiflash.h"
#include "sbfs.h"
#include <dev1000.h>


// Do we want to get debug screen output? It's available if the code
// is loaded with vs3emu (build script) with RS-232 cable.
#if PRINT_VS3EMU_DEBUG_MESSAGES
#define do__not__puts(x) Writeln(x)
#define do__not__puthex(x) puthex(x)
#else
#define do__not__puts(a)
#define do__not__puthex(a)
#define PrintCache()
#endif

u_int32 bytesReceived;

const u_int16 __x charToNum[] = {0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0,0,10,11,12,13,14,15};
const u_int16 __y charToNumH[] = {0*16,1*16,2*16,3*16,4*16,5*16,6*16,7*16,8*16,9*16,0*16,0*16,0*16,0*16,0*16,0*16,0*16,10*16,11*16,12*16,13*16,14*16,15*16};

auto u_int16 ChCh2Hex(register char msb, register char lsb){
	bytesReceived++;
	return ((charToNumH-'0')[msb]) | (charToNum-'0')[lsb];
}

void BusyWait(register u_int16 i){
	while(i--) BusyWait10();
}

//#define USE_LOOPING /*Adds looping, removes Download and Read File commands.*/

#ifdef USE_LOOPING
#include <codec.h>
// ------------ Gapless looping functionality --------------
// Expose some ROM functions...
auto s_int16 CodVorbisGoTo(register struct CodecVorbis *cv,
            register u_int16 seconds);
extern struct Codec *cod;
struct Codec *CodVorbisCreate(void);
extern struct CodecServices cs;
extern u_int16 ogg[];
extern s_int16 vFirstFrame;
// PlayCurrentFile with gapless looping
u_int16 MyPlayCurrentFile(void) {
  register enum CodecError ret = ceFormatNotFound;
  LoadCheck(NULL, 0); // set higher clock
  if ((cod = CodVorbisCreate())) {
    const char *eStr;
    //ret = cod->Decode(cod, &cs, &eStr); // play Ogg Vorbis
    ret = PatchCodVorbisDecode(cod, &cs, &eStr, 0); //patch silent channel??? doesn't seem to work?
    // if played to the end, start over
Writeln("Once.");	
    
    while (cs.fileLeft < 5) {
      LoadCheck(NULL, 0);    // set higher clock
      CodVorbisGoTo(cod, 0); // go to beginning of file
      ogg[1] = 0;            // ogg.headerTypeFlag: reset lastFrame
      ogg[14] = 0;           // ogg.cont: set vorbis continuation
      vFirstFrame = 1;       // first ogg frame in file
      cs.playTimeSamples = 0;
      cs.playTimeSeconds = 0;
      //ret = cod->Decode(cod, &cs, &eStr); // play Ogg Vorbis
      ret = PatchCodVorbisDecode(cod, &cs, &eStr, 0); //patch silent channel
    }
    //cod->Delete(cod); //does not really do anything...
  }
  return ret;
}
#endif

//char *cp;
char filename[112];	

#define buf minifatBuffer


extern struct FsMapper *map;

auto void GetFileName(){
        register char *cp = filename;
	register char ch;
	//Receive file name until '<'
	Writeln("Name:");	
	while (((ch=MyGetChar()) >= ' ') && (ch != '<')){
		putch2(ch);
		*cp++ = ch;
	}
	*cp=0; //Ending zero of pchar filename
	putch2('\r');
}

//#define QUIET_CANCEL /*Volume down at cancel point.*/

#include <codec.h> // CODEC interface
extern struct CodecServices cs;
void VoiceIdleHook(void) {
    if (CharReady()) {
	register int cmd = MyGetChar();
	if (cmd == 'C') { /* Cancel play */
	    Writeln("Cancel");
	    cs.cancel = 1;
	    player.pauseOn = 0;
	} else if (cmd == '+') {
	    KeyEventHandler(ke_volumeUp2);
	} else if (cmd == '-') {
	    KeyEventHandler(ke_volumeDown2);
	} else if (cmd == '=') {
	    /* pause */
	    Writeln("Pause");
	    player.pauseOn = 1;
	} else if (cmd == '>') {
	    /* play */
	    Writeln("Play");
	    player.pauseOn = 0;
	    cs.fastForward = 1;
	} else if (cmd == '') {
	    /* Fast forward */
	    //if (cs.fastForward < 4)
	    Writeln("FF");
	    ++cs.fastForward;
	} else {
	    /*unknown command*/
	}
    }
//    if (uiTrigger) {
//	uiTrigger = 0;
//    }
#ifdef QUIET_CANCEL
    if (cs.cancel) {
	volumeReg = 0xc0c0;
	SetVolume();
    }
#endif
    return;
}


/* Our own loadcheck which will use fixed 3.0x clock for player
   and 4.0x for USB. But as we don't have USB mode in this program,
   we would only need fixed clock. */
void UartLoadCheck(struct CodecServices *cs, s_int16 n) {
  if (cs == NULL) {
#if 0 /* USB is not serviced */
    if (n) {
	clockX = 8; /* USB requires 4.0x! */
	goto set;
    }
#endif
    if (clockX != 6) {
	clockX = 6; /* Otherwise 3.0x */
    set:
	SetRate(hwSampleRate);
    }
    return;
  }
#if 1
  /* vorbis/replaygain - cs is always non-NULL here */
  if (cs->gain != player.volumeOffset) {
    player.volumeOffset = cs->gain; /* vorbis-gain */
    PlayerVolume();
  }
#endif
}

void main(void) {
  u_int16 i;

	voltages[voltIoPlayer] = voltages[voltIoUSB] = 27; /*3.3V*/
	PowerSetVoltages(&voltages[voltCorePlayer]);

	InitAudio();
 	SetRate(cs.sampleRate = 8000);
	// Enable Interrupts
	SetSerialInterruptVector();
	PERIP(INT_ENABLEH) = 0;
	PERIP(INT_ENABLEL) = INTF_RX | INTF_DAC;
	
	// Drop from USB
	PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA;
	//PERIP(USB_CONFIG) = 0x8000U;
  	
	PERIP(GPIO1_ODATA) |=  LED1|LED2;
	PERIP(GPIO1_DDR)   |=  LED1|LED2;
	PERIP(GPIO1_MODE)  &= ~(LED1|LED2);
  	
	player.volumeOffset = 0;
	//keyOld = KEY_POWER;
	//keyOldTime = -32767;
	
//	SetHookFunction((u_int16)OpenFile, SbfsOpenFile);
	SetHookFunction((u_int16)IdleHook, VoiceIdleHook);
	SetHookFunction((u_int16)ReadDiskSector, EeReadDiskSector);
	SetHookFunction((u_int16)USBSuspend, NullHook); /*Disable low-power pause*/
	SetHookFunction((u_int16)LoadCheck, UartLoadCheck); /*fixed clock*/

 	InitSpi();
	goto print_ident; /* start by printing the help screen */

	while(1) {
		char opcode,ch;
		Writeln("\r#");
		opcode = MyGetChar();
		if (opcode >= 'a' && opcode <= 'z') {
		    opcode += 'A'-'a';
		}
		putch2(opcode);
		putch2('\r');
		switch(opcode) {
		case 'F': //Files
			SbfsOpenFile(FILE_LIST,NULL);
			break;
		case 'Q': //Quick
			waitStates = 0;
			Writeln("Quick\r");
			break;
		case 'S': //Slow
			waitStates = 1;
			Writeln("Slow\r");
			break;
		case 'E': //Erase all
		  {
		    register u_int32 bl;
		    Writeln("Erasing"); /*supports only 4k/64k erase.*/
		        for (bl=FIRST_BLOCK; bl<=LAST_BLOCK;){
			    /* Check if erase needed */
			    {
				register u_int16 *p = minifatBuffer;
				register u_int16 d = 0xffff, i;
				EeReadDiskSector(minifatBuffer, bl);
				for (i=0;i<256;i++) {
				    d &= *p++;
				}
				if (d == 0xffff) { /* already erased */
				    bl++;
				    continue;
				}
			    }
			    if ((bl & 127)) { /* not aligned */
//				puthex32(bl);
				EeErase4K(bl);
				bl = (bl & ~(ERASEBLOCK_SIZE-1))
				    + ERASEBLOCK_SIZE;
//				putch2(',');
			    } else { /* aligned to 64k */
//				puthex32(bl);
				EeErase64K(bl);
				bl += (bl & ~127) + 128;
				putch2('.');
			    }
			}
		        Writeln(" Done\r");
			break;
			}
		case 'D': //Delete file
			GetFileName();
			SbfsEraseFile(filename);		
			break;
		case 'P': //Play file
			GetFileName();
			if (SbfsOpenFile(0,filename)){
				Writeln("Play..");
				PlayerVolume(); /*required for QUIET_CANCEL*/
				cs.cancel = 0;
				cs.goTo = -1;
				cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
				cs.fastForward = 1;
#ifdef USE_LOOPING
				MyPlayCurrentFile();
#else
				//PatchPlayCurrentFile2(); //too large
				PatchPlayCurrentFile(); //silent channel fix
				//PlayCurrentFile();
#endif
			} else {
				Writeln("File not found.\r");
			}
			break;
#ifndef USE_LOOPING
		case 'L': //Download file
		case 'R': //Read file
			GetFileName();
			if (SbfsOpenFile(0,filename)){
				char c;
				while (c=0,ReadFile(&c,1,1)) {
					if (opcode=='R') {
						putch2(c);
					} else {
						putch2(hex[c>>4]);
						putch2(hex[c&15]);
					}
				}				
			} else {
				Writeln("File not found.\r");
			}

			break;
#endif
		case 'U': //Upload file
			GetFileName();
			if (!SbfsCreateFile(filename)) {
			        //cp = buf;
				Writeln("Send upcase hex now and end with \".\"");
				BusyWait10();
				putch2('>');
				BusyWait10();
				bytesReceived = 0;
				while(1) {
					u_int16 b;
					if ((ch = MyGetChar()) == '.') {
						goto out;
					}
					while (ch<'0') {
						ch = MyGetChar();
					}
					b=ChCh2Hex(ch,MyGetChar());
					if ((bytesReceived & 0x3ff) == 1) {
						putch2('~');
					}
					SbfsWriteByte(b);
				}
			out:
				SbfsCloseFile(filename);
				putch2('\r');
				puthex32(bytesReceived);
				Writeln("bytes received.\r");
			}
			break;
		case 'H': //Help
		print_ident:
		  //puthex32(LAST_BLOCK);
			Writeln("VS1000 UART Audio Module V1.4\r");
			if (LAST_BLOCK == 0x7fff){
			    Writeln("16MB\r");
			} else if (LAST_BLOCK == 0x1fff) {
			    Writeln("4MB\r");
			} else if (LAST_BLOCK == 0x1ffff) {
			    Writeln("64MB\r");
			} else {
			    /* Otherwise 2MB */
			}
			Writeln("Commands: H)elp F)ilelist U)pload downL)oad R)ead E)raseAll D)eleteFile P)lay\rQ)uick S)low\r"
				"During play:\r '=' pause, '>' play, 'C' stop,\r '' fast forward, '+' vol up, '-' vol down");
			//puthex32(LAST_BLOCK);
			//Writeln(" last block\r");
			break;
		case '\r': // Intentional fall-through
		case '\n':
			break;
		default:
			Writeln("Unknown command");
			break;
		}
	}	
}

